import logging

""" Set log format considering
lavelname:
    DEBUG: 5
    INFO: 4
    WARNING: 7
    ERROR: 5
    CRITICAL: 8
name: module name
"""
LOG_FORMAT = '%(asctime)s %(levelname)-8s %(name)-26s %(message)s'


def configure_logging(verbose):
    """Configure application-wide logging."""

    logging.basicConfig(format=LOG_FORMAT)

    # As logging.basicConfig() does anything only the first time it is
    # called, the root logger should be configured explicitly
    root_logger = logging.getLogger()

    log_level = logging.DEBUG if verbose else logging.INFO
    root_logger.setLevel(log_level)

    log = logging.getLogger(__name__)
    log.debug('Added logging handler to logger root at %s', __name__)



from pathlib import Path

from rpmlb.downloader.local import LocalDownloader


def test_init():
    downloader = LocalDownloader()
    assert downloader


def test_download(tmpdir_factory):

    downloader = LocalDownloader()
    package_dict = {'name': 'a'}

    # Create source files for test.
    src_dir = tmpdir_factory.mktemp('rpmlb-test-src-dir')

    original_spec_path = Path(str(src_dir), 'a', 'a.spec')
    original_spec_path.parent.mkdir(parents=True)
    original_spec_path.touch()

    # "Download" the files
    with tmpdir_factory.mktemp('rpmlb-test-download-dir').as_cwd():
        downloader.download(package_dict, source_directory=str(src_dir))
        spec_path = Path('a', 'a.spec')
        assert spec_path.is_file()



from collections import defaultdict
from collections.abc import Mapping

import pytest

from rpmlb.recipe import Recipe, RecipeError


@pytest.fixture
def ok_recipe(valid_recipe_path):
    return Recipe(str(valid_recipe_path), 'rh-ror50')


@pytest.fixture
def invalid_pkg_type():
    """Package entry of invalid type"""

    return float('nan')  # anything but str or Mapping


@pytest.fixture
def invalid_pkg_content():
    """Package entry with an invalid content"""

    return {'pkg-a': {}, 'pkg-b': {}}


@pytest.fixture
def invalid_recipe(ok_recipe, invalid_pkg_type, invalid_pkg_content):
    """Recipe with invalid package contents"""

    ok_recipe.recipe['packages'] += [
        invalid_pkg_content,
        invalid_pkg_type,
    ]

    return ok_recipe


@pytest.fixture
def bootstrap_recipe(ok_recipe):
    """Recipe with multiple package instances that need bootstrapping"""

    ok_recipe.recipe['packages'] += [
        'pkg-a', 'pkg-b',
        'pkg-a', 'pkg-b',
        'pkg-a', 'pkg-b',
    ]

    return ok_recipe


def test_init(ok_recipe):
    assert ok_recipe


def test_recipe(ok_recipe):
    assert ok_recipe.recipe


def test_recipe_not_found(valid_recipe_path):
    with pytest.raises(KeyError):
        Recipe(str(valid_recipe_path), 'dummy')


def test_verify_returns_true_on_valid_recipe(ok_recipe):
    assert ok_recipe.verify()


def test_verify_returns_true_on_recipe_with_name(ok_recipe):
    ok_recipe.recipe['name'] = 'a'
    assert ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_empty_name(ok_recipe):
    ok_recipe.recipe['name'] = ''
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_invalid_name(ok_recipe):
    ok_recipe.recipe['name'] = ['a']
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_returns_true_on_recipe_with_requires(ok_recipe):
    ok_recipe.recipe['requires'] = [
        'rh-ruby23', 'rh-nodejs4', 'rh-mongodb32',
    ]
    assert ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_non_list_requires(ok_recipe):
    ok_recipe.recipe['requires'] = 'abc'
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_without_packages(ok_recipe):
    del ok_recipe.recipe['packages']
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_empty_packages(ok_recipe):
    ok_recipe.recipe['packages'] = []
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_non_list_packages(ok_recipe):
    ok_recipe.recipe['packages'] = 'abc'
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_empty_package_name(ok_recipe):
    ok_recipe.recipe['packages'] = [
        'foo',
        ''
    ]
    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_on_recipe_with_unknown_element(ok_recipe):
    ok_recipe.recipe['unknown'] = 'a'
    with pytest.raises(RecipeError):
        ok_recipe.verify()


@pytest.mark.parametrize('key', ('macros', 'replaced_macros', 'cmd', 'dist'))
def test_verify_raises_error_with_empty_value_in_package(ok_recipe, key):
    package = {
        'foo_package': {
            key: ''
        }
    }
    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


@pytest.mark.parametrize('key', ('macros', 'replaced_macros'))
def test_verify_raises_error_with_invalid_macros_in_package(ok_recipe, key):
    package = {
        'foo_package': {
            key: ['a']
        }
    }
    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_with_invalid_cmd_in_package(ok_recipe):
    package = {
        'foo_package': {
            'cmd': {
                'a': 'cmd_a',
                'b': 'cmd_b',
            }
        }
    }

    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_with_invalid_dist_in_package(ok_recipe):
    package = {
        'foo_package': {
            'dist': ['el6', 'el7']
        }
    }

    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_with_unknown_element_in_package(ok_recipe):
    package = {
        'foo_package': {
            'unknown': 'a',
        }
    }

    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_with_multi_key_in_package(ok_recipe):
    package = {
        'foo_package': {
            'cmd': 'ls'
        },
        'bar_package': {
            'cmd': 'ls'
        }
    }

    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_raises_error_with_invalid_package_type(ok_recipe):
    package = ['foo_package', 'a']

    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(RecipeError):
        ok_recipe.verify()


def test_verify_recipe_returns_messages_on_error(ok_recipe):
    a_package = {
        'a_package': {
            'macros': 'a',
            'patch': 'ls'
        }
    }
    ok_recipe.recipe['packages'].append(a_package)
    messages = ok_recipe._verify_recipe()
    assert len(messages) == 2


def test_verify_recipe_returns_empty_messages_on_success(ok_recipe):
    a_package = {
        'a_package': {
            'macros': {
                'a': 'A'
            },
            'dist': 'fc23'
        }
    }
    ok_recipe.recipe['packages'].append(a_package)
    messages = ok_recipe._verify_recipe()
    assert len(messages) == 0


def test_verify_packages_raises_error_on_messages_none(ok_recipe):
    messages = None
    packages = ['a']
    with pytest.raises(ValueError):
        ok_recipe._verify_packages_and_append(messages, packages)


def test_verify_packages_returns_on_empty_messages(ok_recipe):
    messages = []
    packages = ['a']
    out_messages = ok_recipe._verify_packages_and_append(messages, packages)
    assert len(out_messages) == 0


def test_verify_packages_raises_error_on_packages_none(ok_recipe):
    messages = ['a']
    packages = None
    with pytest.raises(ValueError):
        ok_recipe._verify_packages_and_append(messages, packages)


def test_verify_packages_metadata_raises_error_on_messages_none(ok_recipe):
    messages = None
    metadata = ['a']
    disp_package = 'a'
    with pytest.raises(ValueError):
        ok_recipe._verify_packages_metadata_and_append(
            messages, metadata, disp_package)


def test_verify_packages_metadata_raises_error_on_metadata_none(ok_recipe):
    messages = ['a']
    metadata = None
    disp_package = 'a'
    with pytest.raises(ValueError):
        ok_recipe._verify_packages_metadata_and_append(
            messages, metadata, disp_package)


def test_verify_packages_metadata_raises_error_on_disp_package_none(ok_recipe):
    messages = ['a']
    metadata = {
        'cmd': 'ls'
    }
    disp_package = None
    with pytest.raises(ValueError):
        ok_recipe._verify_packages_metadata_and_append(
            messages, metadata, disp_package)


def test_package_name_type_checks(ok_recipe):
    """Invalid package types are reported."""

    invalid_type = float('nan')
    with pytest.raises(ValueError):
        ok_recipe._package_name(invalid_type)

    valid_type = 'pkg-a'
    name = ok_recipe._package_name(valid_type)
    assert name == 'pkg-a'

    invalid_contents = {'pkg-a': {}, 'pkg-b': {}}
    with pytest.raises(ValueError):
        ok_recipe._package_name(invalid_contents)

    valid_contents = {'pkg-a': {}}
    name = ok_recipe._package_name(valid_contents)
    assert name == 'pkg-a'


def test_package_normalization_type_checks(invalid_recipe):
    """Invalid packages are reported when normalizing."""

    with pytest.raises(ValueError):
        list(invalid_recipe.each_normalized_package())


def test_package_normalization_raises_error_on_invalid_dict(ok_recipe):
    package = {
        'a_package': {
            'macros': {
                'a': 'A'
            }
        },
        'b_package': {
            'macros': {
                'a': 'A'
            }
        }
    }
    ok_recipe.recipe['packages'].append(package)

    with pytest.raises(ValueError):
        list(ok_recipe.each_normalized_package())


def test_normalized_packages_have_expected_format(ok_recipe):
    """After normalization, all packages have expected format."""

    a_package = {
        'a_package': {
            'macros': {
                'a': 'A'
            },
            'replaced_macros': {
                'b': 'B'
            },
        }
    }
    ok_recipe.recipe['packages'].append(a_package)

    def valid_item(package: Mapping, key: str, types: tuple):
        """Check validity of a package metadata item."""

        exists = key in package
        value = package.get(key, None)
        valid_type = isinstance(value, types)

        return exists and valid_type

    all_packages = list(ok_recipe.each_normalized_package())
    packages_with_macros = [pkg for pkg in all_packages if 'macros' in pkg]
    packages_with_replaced_macros = [
        pkg for pkg in all_packages if 'replaced_macros' in pkg
    ]

    # Each package has a name
    assert all('name' in pkg for pkg in all_packages), all_packages

    # Each package has bootstrap position of correct type
    key_types = 'bootstrap_position', (type(None), int)
    assert all(valid_item(pkg, *key_types) for pkg in all_packages), \
        all_packages

    # Each macros value is a mapping
    key_types = 'macros', (Mapping,)
    assert all(valid_item(pkg, *key_types) for pkg in packages_with_macros), \
        packages_with_macros

    # Each replaced_macros value is a mapping
    key_types = 'replaced_macros', (Mapping,)
    assert all(
        valid_item(pkg, *key_types)
        for pkg in packages_with_replaced_macros
    ), packages_with_replaced_macros


def test_bootstrap_sequence_increments_correctly(bootstrap_recipe):
    """Packages that needs bootstrapping are correctly marked as such."""

    sequences = defaultdict(list)
    for pkg in bootstrap_recipe.each_normalized_package():
        sequences[pkg['name']].append(pkg['bootstrap_position'])

    assert sequences['pkg-a'] == [1, 2, None], sequences['pkg-a']



import logging

from rpmlb.downloader.base import BaseDownloader

LOG = logging.getLogger(__name__)


class NoneDownloader(BaseDownloader):
    """A downloader class to do nothing."""

    def download(self, package_dict, **kwargs):
        pass



from pathlib import Path
from subprocess import STDOUT, check_output

# These tests use subprocess directly due to weird interactions with test_cli
# that lead to the command output being unreliable when using click.CliRunner.

_TEST_DIR = Path(__file__).parent.parent
_CUSTOM_HOOK_DIR = _TEST_DIR / "fixtures" / "custom"
_CUSTOM_HOOKS = _CUSTOM_HOOK_DIR / "echo.yml"
_EXAMPLE_RECIPE = _TEST_DIR / "fixtures" / "recipes" / "echo_recipe.yml"

# The custom echo backend is fast enough & simple enough that we just run a
# full end-to-end integration test as the custom builder's unit tests
#
# Both the download & build hooks need to be run, or the latter fail due to
# the missing package directories

_cli_command = [
    "rpmlb", "--custom-file", str(_CUSTOM_HOOKS),
    "--download", "custom", "--build", "custom",
    str(_EXAMPLE_RECIPE), "echo-packages",
]


def test_custom_build_hooks():
    """Test that `--build custom` works as expected"""
    output = check_output(_cli_command, universal_newlines=True, stderr=STDOUT)
    lines = output.strip().splitlines()
    # Only a unit test, so we don't explicitly check hook order,
    # just the expected implicit environment variable declarations
    hook_dir_line = "Using hook dir: {}".format(_CUSTOM_HOOK_DIR)
    package_1_build_line = "Building PKG: a-package"
    package_2_build_line = "Building PKG: another-package"
    assert hook_dir_line in lines
    assert package_1_build_line in lines
    assert package_2_build_line in lines



import logging
import os
import shutil
import tempfile

from rpmlb import utils

LOG = logging.getLogger(__name__)


class Work:
    """A class to manage working directory."""

    def __init__(self, recipe, **kwargs):
        if recipe is None:
            raise ValueError('recipe is required.')

        self._recipe = recipe
        max_digit = len(str(recipe.num_of_package))
        self._num_dir_format = '%0{0}d'.format(str(max_digit))

        working_dir = None
        if 'work_directory' in kwargs and kwargs['work_directory']:
            working_dir = kwargs['work_directory']
            if not os.path.exists(working_dir):
                os.makedirs(working_dir)
        else:
            working_dir = tempfile.mkdtemp(prefix='rpmlb-')
        LOG.info('Working directory: %s', working_dir)
        self.working_dir = working_dir

    def close(self):
        if os.path.isdir(self.working_dir):
            shutil.rmtree(self.working_dir)

    def num_name_from_count(self, count):
        num_name = self._num_dir_format % count
        return num_name

    def each_num_dir(self):
        if not os.path.isdir(self.working_dir):
            ValueError('working_dir does not exist.')

        count = 1
        for package_dict in self._recipe.each_normalized_package():
            num_name = self.num_name_from_count(count)
            num_dir = os.path.join(self.working_dir, num_name)

            if not os.path.isdir(num_dir):
                os.makedirs(num_dir)
            with utils.pushd(num_dir):
                yield package_dict, num_name

            count += 1

    def each_package_dir(self):
        for package_dict, num_name in self.each_num_dir():
            package = package_dict['name']
            with utils.pushd(package):
                yield package_dict, num_name



"""Test configuration for py.test."""

import os
import random
import sys
import tempfile
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Union

import pytest

pytest_plugins = ['helpers_namespace']

sys.path.append(os.path.join(os.path.dirname(__file__)))


# Fixture configuration

#: Number of bits in generated random numbers
RANDOM_BITS = 32
#: Common prefix of generated temporary files
TMP_FILE_PREFIX = 'rpmlb-tmp-'


# Helper functions

@pytest.helpers.register
@contextmanager
def pushd(target_dir: Union[str, Path]):
    """Change current directory to target_dir."""

    current_dir = os.getcwd()
    if isinstance(target_dir, Path):
        target_dir = str(target_dir)

    try:
        os.chdir(target_dir)
        yield
    finally:
        os.chdir(current_dir)


@pytest.helpers.register
def remove_if_is_file(path: Union[str, Path]):
    """Remove file if it exists."""

    if not isinstance(path, Path):
        path = Path(path)

    if path.is_file():
        path.unlink()


@pytest.helpers.register
@contextmanager
def generate_tmp_path():
    """Generate randomly named temporary path.

    The file is removed at the exit from the context, if it exists.
    """

    tmp_path = Path(tempfile.gettempdir(), '{prefix}{suffix:x}'.format(
        prefix=TMP_FILE_PREFIX,
        suffix=random.getrandbits(RANDOM_BITS),
    ))

    try:
        yield tmp_path
    finally:
        remove_if_is_file(tmp_path)


@pytest.helpers.register
@contextmanager
def generate_tmp_path_list(count: int):
    """Generate multiple temporary paths.

    All files are removed at the exit from the context, if they exist.

    Keyword arguments:
        count: Number of paths to generate.
    """

    with ExitStack() as stack:
        tmp_path_list = [
            stack.enter_context(generate_tmp_path()) for __ in range(count)
        ]

        yield tmp_path_list


# Fixtures

@pytest.fixture
def random_file_path():
    """Provide path to a random temporary file."""

    with generate_tmp_path() as path:
        yield path


@pytest.fixture(scope='session')
def valid_recipe_path():
    """Provide path to a valid recipe file."""

    return Path('tests/fixtures/recipes/ror.yml')


@pytest.fixture(scope='session')
def valid_custom_file_path():
    """Provide path to a valid custom file."""

    return Path('tests/fixtures/custom/echo.yml')



import logging
import os
import shutil

from rpmlb.downloader.base import BaseDownloader

LOG = logging.getLogger(__name__)


class LocalDownloader(BaseDownloader):
    """A downloader class to copy a pacakge from source directory."""

    def __init__(self):
        pass

    def download(self, package_dict, **kwargs):
        if not package_dict:
            raise ValueError('package_dict is required.')
        if 'source_directory' not in kwargs or not kwargs['source_directory']:
            raise ValueError('source directory is required.')
        src_dir = kwargs['source_directory']
        package = package_dict['name']
        src_package_dir = os.path.join(src_dir, package)
        dst_package_dir = os.path.join(os.getcwd(), package)
        LOG.debug('Copying %s to %s .', src_package_dir, dst_package_dir)

        def ignore_symlinks(directory, files):
            ignored_files = []
            for f in files:
                full_path = os.path.join(directory, f)
                if os.path.islink(full_path):
                    ignored_files.append(f)
            return ignored_files

        # Skip symbolic file when copying because the file is risky.
        shutil.copytree(
            src_package_dir,
            dst_package_dir,
            ignore=ignore_symlinks,
            symlinks=True
        )



import logging
import os

import yaml

from . import utils

LOG = logging.getLogger(__name__)


class Yaml:
    """A class to manage YAML data."""

    def __init__(self, file_path):
        try:
            from yaml import CLoader as Loader
        except ImportError:
            from yaml import Loader
        try:
            with open(str(file_path), 'r') as stream:
                self.content = yaml.load(stream, Loader=Loader)
        except FileNotFoundError as e:
            LOG.error('File not found: %s at %s', file_path, os.getcwd())
            raise e

    def dump(self):
        try:
            from yaml import CDumper as Dumper
        except ImportError:
            from yaml import Dumper
        output = yaml.dump(self.content, Dumper=Dumper,
                           default_flow_style=False)
        print(output)

    @staticmethod
    def run_cmd_element(cmd_element, **kwargs):
        if isinstance(cmd_element, str):
            cmds = [cmd_element]
        elif isinstance(cmd_element, list):
            cmds = cmd_element
        else:
            message = 'Invalid type of cmd_element: {!r}'.format(
                type(cmd_element)
            )
            raise ValueError(message)
        for cmd in cmds:
            utils.run_cmd(cmd, **kwargs)



import re
import sys
from collections import Counter
from pathlib import Path
from unittest import mock

import pytest

from rpmlb.builder.base import MACRO_REGEX, BaseBuilder
from rpmlb.recipe import Recipe
from rpmlb.work import Work


@pytest.fixture
def empty_spec_path(tmpdir):
    """Empty spec file path"""

    spec = Path(str(tmpdir), 'test.spec')
    spec.touch()

    with tmpdir.as_cwd():
        yield spec

    if spec.exists():
        spec.unlink()


@pytest.fixture
def prepared_macros():
    """Prepared testing spec macros"""

    macro_dict = {
        # Simple macro value
        'test_macro_a': r'value',
        # Macros with white space in the body
        'test_macro_b': r'with spaces in value',
        # Multi-line macro
        'test_macro_c': r'''multi \
            line''',
    }

    return macro_dict


@pytest.fixture
def macro_spec_path(empty_spec_path, prepared_macros):
    """Path to a spec file containing some macro definitions"""

    with empty_spec_path.open(mode='w') as out_file:
        for name, value in prepared_macros.items():
            print('%global {} {}'.format(name, value), file=out_file)

        print('BuildRequires foo', file=out_file)
        print('BuildRequires bar', file=out_file)
        print('BuildRequires dist', file=out_file)

    return empty_spec_path


@pytest.fixture
def builder(valid_recipe_path):
    """Provide initialized BaseBuilder."""

    recipe = Recipe(str(valid_recipe_path), 'rh-ror50')
    work = Work(recipe)
    builder = BaseBuilder(work, pkg_cmd='testpkg')
    return builder


def test_init(builder):
    assert builder


def test_init_raises_error_without_pkg_cmd_option(valid_recipe_path):
    recipe = Recipe(str(valid_recipe_path), 'rh-ror50')
    work = Work(recipe)
    with pytest.raises(ValueError):
        BaseBuilder(work)


def test_run_calls_before_build_after(builder):
    builder.before = mock.MagicMock()
    builder.build = mock.MagicMock(return_value=True)
    builder.after = mock.MagicMock()
    builder.prepare = mock.MagicMock()

    mock_work = get_mock_work()
    builder.run(mock_work)
    assert builder.before.called
    assert builder.build.called
    assert builder.after.called


def test_run_returns_true_on_success(builder):
    builder.build = lambda *args, **kwargs: None
    builder.prepare = mock.MagicMock()

    mock_work = get_mock_work()
    assert builder.run(mock_work)


def test_run_exception(builder):
    builder.build = mock.Mock(side_effect=ValueError('test'))
    builder.prepare = mock.MagicMock()

    mock_work = get_mock_work()
    type(mock_work).working_dir = mock.PropertyMock(return_value='work_dir')

    called = False
    error_message = None
    try:
        builder.run(mock_work, retry=3)
    except RuntimeError as e:
        called = True
        error_message = str(e)
    assert called
    expected_error_message = (
        "pacakge_dict: {'name': 'a'}, "
        "num: 1, work_dir: work_dir"
    )
    assert error_message == expected_error_message
    if sys.version_info >= (3, 6):
        builder.build.assert_called()
    # bulid should be called 3 times for retry setting.
    calls = [mock.call({'name': 'a'}, retry=3)] * 4
    builder.build.assert_has_calls(calls)
    assert builder.build.call_count == len(calls)


def test_run_does_not_call_before_calls_build_after_with_resume_option(builder):  # noqa: E501
    builder.before = mock.MagicMock()
    builder.build = mock.MagicMock(return_value=True)
    builder.after = mock.MagicMock()
    builder.prepare = mock.MagicMock()

    mock_work = get_mock_work()
    builder.run(mock_work, resume=2)
    assert not builder.before.called
    assert builder.build.called
    assert builder.after.called


def get_mock_work():
    mock_work = mock.MagicMock()
    package_dicts = [
        {'name': 'a'},
        {'name': 'b'},
    ]
    num_names = [
        '1',
        '2',
    ]
    mock_work.each_package_dir.return_value = iter(
        zip(package_dicts, num_names))
    return mock_work


def test_double_edit(empty_spec_path):
    """Original spec file is not modified twice"""

    def edit_file(path, indicator='Test edit'):
        """Single edit round"""

        with BaseBuilder.edit_spec_file(str(path)) as (src, dst):
            print(indicator, file=dst)
            dst.write(src.read())

    def count_markers(path, marker='# Edited by rpmlb'):
        """Count lines containing marker text in file"""

        with path.open() as handle:
            return sum(marker in line for line in handle)

    for indicator in ('First edit', 'Second edit'):
        edit_file(empty_spec_path, indicator)
        assert empty_spec_path.exists()
        assert count_markers(empty_spec_path) == 1


def test_macros_are_added_correctly(macro_spec_path, prepared_macros):
    """Macros are safely added without messing with current file contents."""

    new_macro_dict = {
        'bootstrap': 1,
        'with_bootstrap': 1,
    }

    with BaseBuilder.edit_spec_file(macro_spec_path) as (original, modified):
        content_stream = BaseBuilder.add_macros(original, new_macro_dict)
        modified.write(''.join(content_stream))

    with macro_spec_path.open() as spec_file:
        spec_contents = spec_file.read()

    # Count the macro occurrences
    matches = MACRO_REGEX.finditer(spec_contents)
    macro_counts = Counter(match.group('name') for match in matches)

    # Each expected macro is present exactly once
    assert all(cnt == 1 for cnt in macro_counts.values()), macro_counts

    expected_macro_names = new_macro_dict.keys() | prepared_macros.keys()
    present_macro_names = macro_counts.keys()
    assert expected_macro_names == present_macro_names, present_macro_names


def test_macros_are_replaced_correctly(macro_spec_path, prepared_macros):
    """Macros are correctly replaced."""

    replaced_macro_dict = dict.fromkeys(prepared_macros, 'REPLACED')

    with BaseBuilder.edit_spec_file(macro_spec_path) as (original, modified):
        content_stream = BaseBuilder.replace_macros(
            original,
            replaced_macro_dict,
        )
        modified.write(''.join(content_stream))

    with macro_spec_path.open() as spec_file:
        spec_contents = spec_file.read()

    result_macros = dict(MACRO_REGEX.findall(spec_contents))

    # All expected macros have their contents replaced
    assert all(val == 'REPLACED' for val in result_macros.values()), \
        result_macros  # show the macro values when debugging


def test_prepare_runs_all_preparations(
    builder, macro_spec_path, prepared_macros,
):
    """All preparations are done as expected

    1. `replace_macros` are replaced
    2. New macros are added
    3. Any builder extra steps are executed
    4. Commands in cmd element are executed
    """

    spec_file_name = macro_spec_path.name
    package_metadata = {
        'name': macro_spec_path.stem,
        'macros': dict(a='macro a', b='macro b'),
        'replaced_macros': dict.fromkeys(prepared_macros, 'REPLACED'),
        'cmd': [
            'sed -i "/^BuildRequires foo$/ s/^/#/" {0}'.format(spec_file_name),
            'sed -i "/^BuildRequires bar$/ s/^/#/" {0}'.format(spec_file_name),
            r'''if [[ "${{DIST}}" =~ centos ]]; then
                 sed -i "/^BuildRequires dist$/ s/dist/${{DIST}}/" {0}
             fi
            '''.format(spec_file_name),
        ],
    }

    builder.prepare_extra_steps = mock.MagicMock(
        wraps=builder.prepare_extra_steps,
    )

    builder.prepare(package_metadata, dist='centos7')

    with macro_spec_path.open() as spec_file:
        spec_contents = spec_file.read()

    macros = dict(MACRO_REGEX.findall(spec_contents))

    # Adding is performed
    assert all(macros[name] == package_metadata['macros'][name]
               for name in package_metadata['macros'])
    # Replacement is performed
    assert all(macros[name] == 'REPLACED'
               for name in package_metadata['replaced_macros'])
    # Extra preparation steps are performed
    assert builder.prepare_extra_steps.called

    assert re.search('^#BuildRequires foo$',
                     spec_contents, flags=re.MULTILINE)
    assert re.search('^#BuildRequires bar$',
                     spec_contents, flags=re.MULTILINE)
    assert re.search('^BuildRequires centos7$',
                     spec_contents, flags=re.MULTILINE)


def test_is_build_skipped_raises_error_when_package_dict_is_none(builder):
    package_dict = None
    num_name = '1'
    is_redume = False
    with pytest.raises(ValueError):
        builder._is_build_skipped(package_dict, num_name, is_redume)


def test_is_build_skipped_raises_error_when_num_name_is_none(builder):
    package_dict = {'name': 'a'}
    num_name = None
    is_redume = False
    with pytest.raises(ValueError):
        builder._is_build_skipped(package_dict, num_name, is_redume)


def test_is_build_skipped_returns_false_on_matched_dist(builder):
    package_dict = {
        'name': 'a',
        'dist': 'fc2[56]',
    }
    num_name = '1'
    is_redume = False
    kwargs = {
        'dist': 'fc26',
    }
    assert not builder._is_build_skipped(
        package_dict, num_name, is_redume, **kwargs)


def test_is_build_skipped_returns_true_on_unmatched_dist(builder):
    package_dict = {
        'name': 'a',
        'dist': 'fc2[56]',
    }
    num_name = '1'
    is_redume = False
    kwargs = {
        'dist': 'fc2',
    }
    assert builder._is_build_skipped(
        package_dict, num_name, is_redume, **kwargs)



import logging

from rpmlb.builder.base import BaseBuilder

LOG = logging.getLogger(__name__)


class DummyBuilder(BaseBuilder):
    """A dummy builder class."""

    def build(self, package_dict, **kwargs):
        LOG.info('dummy build %s', package_dict['name'])



import pytest

from rpmlb.yaml import Yaml


@pytest.fixture
def valid_yaml(valid_recipe_path):
    """A Yaml object with a valid YAML file."""
    return Yaml(str(valid_recipe_path))


def test_init_loads_file(valid_yaml):
    assert valid_yaml


def test_init_raises_error_on_not_existing_file():
    with pytest.raises(FileNotFoundError):
        Yaml('foo/dummy.yml')


def test_dump_prints_output(capsys, valid_yaml):
    valid_yaml.dump()
    stdout, stderr = capsys.readouterr()
    assert 'rh-ror50:' in stdout


def test_run_cmds_runs_cmd_by_cmd_element_type_string(random_file_path):
    # Test with generated file.
    # Because we can not test using stdout with capsys fixture,
    # the capsys does not capture stdout from subprocess in run_cmds.
    cmd_element = "touch {0!s}".format(random_file_path)
    Yaml.run_cmd_element(cmd_element)
    assert random_file_path.is_file()


def test_run_cmds_runs_cmds_by_cmd_element_type_list():
    with pytest.helpers.generate_tmp_path_list(2) as tmp_files:
        random_file_foo, random_file_bar_part = tmp_files
        random_file_bar = random_file_bar_part.with_name(
            random_file_bar_part.name + '-bar'
        )

        cmd_element = [
            'touch {0}'.format(random_file_foo),
            'touch "{0}-$BAR"'.format(random_file_bar_part),
        ]
        env = {
            'BAR': 'bar'
        }
        Yaml.run_cmd_element(cmd_element, env=env)

        assert random_file_foo.is_file()
        assert random_file_bar.is_file()


def test_run_cmds_raises_error_on_cmd_element_type_dict():
    cmd_element = {
        'a': 'b'
    }
    with pytest.raises(ValueError):
        Yaml.run_cmd_element(cmd_element)



import os
import subprocess
import sys

from rpmlb import utils


def test_camelize():
    new_word = utils.camelize('foo_bar')
    assert new_word == 'FooBar'


def test_get_class():
    cls = utils.get_class('rpmlb.recipe.Recipe')
    assert cls
    cls = utils.get_class(
        'rpmlb.downloader.local.LocalDownloader'
    )
    assert cls


def test_pushd():
    original_dir = os.getcwd()
    with utils.pushd('/tmp'):
        new_dir = os.getcwd()
        assert new_dir == '/tmp'
    current_dir = os.getcwd()
    assert current_dir == original_dir


def test_run_cmd():
    result = utils.run_cmd_with_capture('echo a')
    assert isinstance(result, utils.CompletedProcess)
    assert result.returncode == 0
    assert result.stdout == b'a\n'
    assert result.stderr == b''


def test_run_cmd_exception():
    exception = False
    result_e = None
    cmd = 'ls abc'
    try:
        utils.run_cmd_with_capture(cmd)
    except subprocess.CalledProcessError as e:
        exception = True
        result_e = e

    assert exception
    assert isinstance(result_e, subprocess.CalledProcessError)
    assert result_e.cmd == cmd
    assert result_e.returncode != 0
    assert result_e.output == b''
    if sys.version_info >= (3, 5):
        assert result_e.stdout == b''
        assert b'No such file or directory' in result_e.stderr



from unittest import mock

import pytest

from rpmlb.downloader.base_rpkg import BaseRpkgDownloader


@pytest.fixture
def downloader():
    return BaseRpkgDownloader()


@pytest.fixture
def mock_downloader(downloader, monkeypatch):
    monkeypatch.setattr(type(downloader), 'command',
                        mock.PropertyMock(return_value='testpkg'))

    return downloader


def test_init(downloader):
    assert downloader


def test_command_raises_not_implemented_error(downloader):
    with pytest.raises(NotImplementedError):
        downloader.command


def test_download_passes_on_valid_arguments(mock_downloader):
    package_dict = {'name': 'a'}
    branch = 'private-foo'
    with mock.patch('rpmlb.utils.run_cmd',
                    mock.Mock(return_value=True)) as mock_run_cmd:
        mock_downloader.download(package_dict, branch=branch)
        cmd = r'''
testpkg co a && \
cd a && \
git checkout private-foo
        '''.strip()
        mock_run_cmd.assert_called_once_with(cmd)
    assert True


def test_download_raises_error_on_without_package_dict(downloader):
    package_dict = None
    branch = 'private-foo'
    with mock.patch('rpmlb.utils.run_cmd', mock.Mock(return_value=True)):
        with pytest.raises(ValueError):
            downloader.download(package_dict, branch=branch)


def test_download_raises_error_on_without_branch(downloader):
    package_dict = {'name': 'a'}
    with pytest.raises(ValueError):
        downloader.download(package_dict)



import importlib
import logging
import os
import subprocess
import sys
from contextlib import contextmanager

LOG = logging.getLogger(__name__)


def p(text):
    print(repr(text))


def camelize(word):
    return ''.join(x.capitalize() or '_' for x in word.split('_'))


def get_class(class_name: str, package: str = __package__):
    """Dynamically import class from a module.

    Keyword arguments:
        class_name: The class/attribute name in standard import form.
        package: Name of the package to import from (for relative imports).

    Returns:
        type: The requested class.
    """

    mod_name, class_name = class_name.rsplit(sep='.', maxsplit=1)
    module = importlib.import_module(mod_name, package)
    return getattr(module, class_name)


def get_instance(class_name: str, *args, **kwargs):
    """Dynamically instantiate a class.

    Keyword arguments:
        class_name: Absolute or relative import name of the class.
        *args: Positional arguments for the class __init__.
        **kwargs: Keyword arguments for the class __init__.

    Return:
        Instance of the specified class.
    """

    cls = get_class(class_name)
    return cls(*args, **kwargs)


@contextmanager
def pushd(new_dir):
    previous_dir = os.getcwd()
    try:
        new_ab_dir = None
        if not os.path.isabs(new_dir):
            new_ab_dir = os.path.join(previous_dir, new_dir)
        else:
            new_ab_dir = new_dir
        # Use absolute path to show it on FileNotFoundError message.
        os.chdir(new_ab_dir)
        yield
    finally:
        os.chdir(previous_dir)


def run_cmd_with_capture(cmd, **kwargs):
    kwargs['stdout'] = subprocess.PIPE
    kwargs['stderr'] = subprocess.PIPE
    return run_cmd(cmd, **kwargs)


def run_cmd(cmd, **kwargs):
    returncode = None
    stdout = ''
    stderr = ''
    proc = None
    try:
        check = True
        if 'check' in kwargs:
            check = kwargs['check']
            kwargs.pop('check', None)
        # Use shell option to use wildcard "*".
        kwargs['shell'] = True

        LOG.debug('CMD: %s, kwargs: %s', cmd, repr(kwargs))

        env = os.environ.copy()
        env['LC_ALL'] = 'en_US.utf-8'  # better to parse English output
        if 'env' in kwargs:
            env.update(kwargs['env'])
        kwargs['env'] = env

        proc = subprocess.Popen(cmd, **kwargs)
        stdout, stderr = proc.communicate()
        returncode = proc.returncode
        if check and returncode != 0:
            LOG.error('CMD: [%s] failed at [%s]', cmd, os.getcwd())
            LOG.error('Return Code: %s', returncode)
            if stdout is not None:
                LOG.error('Stdout: %s', stdout)
            if stderr is not None:
                LOG.error('Stderr: %s', stderr)

            kwargs_dict = {}
            kwargs_dict['output'] = stdout
            if sys.version_info >= (3, 5):
                kwargs_dict['stderr'] = stderr
            raise subprocess.CalledProcessError(
                returncode, cmd, **kwargs_dict
            )
        return CompletedProcess(cmd, returncode, stdout, stderr)
    except Exception as e:
        try:
            proc.kill()
        except Exception:
            pass
        raise e


class CompletedProcess:
    """A error class to manage the result of command
    Use it instead of subprocess.CompletedProcess/CalledProcessError
    for old Pytyons (<= 3.4).
    """

    def __init__(self, cmd, returncode, stdout, stderr):
        self.cmd = cmd
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr



import logging
import os
import re

from .. import utils

LOG = logging.getLogger(__name__)


class BaseDownloader:
    """A base class for the package downloader."""

    def __init__(self):
        pass

    @staticmethod
    def get_instance(name: str):
        """Dynamically instantiate named downloader.

        Keyword arguments:
            name: Name of the requested builder.

        Returns:
            Instance of the requested builder.
        """

        class_name = 'rpmlb.downloader.{0}.{1}Downloader'.format(
            name,
            utils.camelize(name)
        )
        instance = utils.get_instance(class_name)

        LOG.debug('Loaded downloader with %s', name)
        return instance

    def run(self, work, **kwargs):
        is_resume = kwargs.get('resume', False)
        if is_resume:
            message = (
                'Skip the download process, '
                'because the resume option was used.'
            )
            LOG.info(message)
            return True

        self.before(work, **kwargs)

        for package_dict, num_name in work.each_num_dir():
            if self._is_download_skipped(package_dict, **kwargs):
                name = package_dict['name']
                LOG.debug('Skip download package: %s', name)
                # Create package directory for build process.
                if not os.path.isdir(name):
                    os.mkdir(name)
                continue

            # TODO(Run it with asynchronous)
            self.download(package_dict, **kwargs)

        self.after(work, **kwargs)
        return True

    def before(self, work, **kwargs):
        pass

    def after(self, work, **kwargs):
        pass

    def download(self, package_dict, **kwargs):
        raise NotImplementedError('Implement this method.')

    def _is_download_skipped(self, package_dict: dict, **kwargs):
        """Return if skip a download for a pacakge.

        Override if needed.

        Keyword arguments:
            package_dict: A dictionary of package metadata.
            kwargs: option arguments.
        """

        if not package_dict:
            raise ValueError('package_dict is required.')

        is_skipped = False

        dist = kwargs.get('dist')
        if dist and 'dist' in package_dict:
            if not re.match(package_dict['dist'], dist):
                is_skipped = True
        return is_skipped



import os
import re
import shutil
import tempfile
from contextlib import closing
from unittest import mock

import pytest

from rpmlb.work import Work


def test_init():
    work = None
    mock_recipe = mock.MagicMock()
    mock_recipe.num_of_package.return_value = 2
    try:
        work = Work(mock_recipe)
        assert work
    finally:
        working_dir = work.working_dir
        assert working_dir
        assert re.findall(r'/tmp/', working_dir)
        if os.path.isdir(working_dir):
            shutil.rmtree(working_dir)


def test_init_work_directory():
    work = None
    arg_working_dir = None
    mock_recipe = mock.MagicMock()
    mock_recipe.num_of_package.return_value = 2
    try:
        arg_working_dir = tempfile.mkdtemp(prefix='rpmlb-test-')
        work = Work(mock_recipe, work_directory=arg_working_dir)
        assert work
    finally:
        working_dir = work.working_dir
        assert working_dir
        assert working_dir == arg_working_dir
        if os.path.isdir(working_dir):
            shutil.rmtree(working_dir)


def test_close():
    mock_recipe = mock.MagicMock()
    type(mock_recipe).num_of_package = mock.PropertyMock(return_value=2)
    work = Work(mock_recipe)
    assert work
    work.close()
    assert not os.path.isdir(work.working_dir)


def test_num_name_from_count():
    mock_recipe = mock.MagicMock()
    type(mock_recipe).num_of_package = mock.PropertyMock(return_value=11)

    with closing(Work(mock_recipe)) as work:
        assert work
        assert work.num_name_from_count(2) == '02'
        assert work.num_name_from_count(10) == '10'


def test_each_num_dir():
    mock_recipe = mock.MagicMock()
    type(mock_recipe).num_of_package = mock.PropertyMock(return_value=2)
    package_dicts = [
        {'name': 'a'},
        {'name': 'b'},
    ]
    mock_recipe.each_normalized_package.return_value = iter(package_dicts)

    with closing(Work(mock_recipe)) as work:
        assert work

        for package_dict, num_name in work.each_num_dir():
            assert package_dict
            assert num_name
            assert re.match('^[12]$', num_name)
            assert re.match('^.+/[12]$', os.getcwd())

        with pytest.helpers.pushd(work.working_dir):
            assert os.path.isdir('1')
            assert os.path.isdir('2')



import logging

from ..custom import Custom
from ..work import Work
from .base import BaseBuilder

LOG = logging.getLogger(__name__)


class CustomBuilder(BaseBuilder):
    """A custom builder class."""

    def __init__(self, work: Work, custom_file: str = None, **options):
        """Initialize the builder.

        Keyword arguments:
            work: The overview of the work to do.
            custom_file: Path to the custom file.
        """

        super().__init__(work, custom_file=custom_file, **options)

        if custom_file is None:
            raise ValueError('custom_file is required.')

        #: Custom commands runner to use
        self.custom_runner = Custom(custom_file)

    def before(self, work, **kwargs):
        self.custom_runner.run_cmds('before_build')

    def build(self, package_dict, **kwargs):
        self.custom_runner.run_cmds('build', **package_dict)



import logging

from rpmlb.custom import Custom
from rpmlb.downloader.base import BaseDownloader

LOG = logging.getLogger(__name__)


class CustomDownloader(BaseDownloader):
    """A custom downloader class."""

    def __init__(self):
        self._yaml_content = None

    def before(self, work, **kwargs):
        custom_file = kwargs['custom_file']
        if not custom_file:
            raise ValueError('custom_file is required.')

        custom = Custom(custom_file)
        custom.run_cmds('before_download')

    def download(self, package_dict, **kwargs):
        custom_file = kwargs['custom_file']
        if not custom_file:
            raise ValueError('custom_file is required.')

        custom = Custom(custom_file)
        custom.run_cmds('download', **package_dict)



import logging
import re
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Iterator, Mapping, Match

from retry.api import retry_call

from .. import utils
from ..work import Work
from ..yaml import Yaml

LOG = logging.getLogger(__name__)


#: Regular expression for finding macro definitions
MACRO_REGEX = re.compile(
    r'''^
    %global\s+                   # beginning of definition
    (?P<name>[^\s]+)\s+          # one-word name
    (?P<value>(?:.|(?<=\\)\n)+)  # value including spaces and escaped newlines
    $''',
    flags=re.MULTILINE | re.VERBOSE,
)


class BaseBuilder:
    """A base class for the package builder."""

    def __init__(self, work: Work, **options):
        """Initialize the builder to a valid state.

        Keyword arguments:
            work: Overview of the work to do.
            options: Command line options for the builder to process.
        """
        if 'pkg_cmd' not in options:
            raise ValueError('pkg_cmd is required.')
        self._pkg_cmd = options['pkg_cmd']

    @staticmethod
    def get_instance(name: str, work: Work, **options):
        """Instantiate named builder.

        Keyword arguments:
            name: The name of the requested builder.
            work: Overview of the work to do.
            options: Command line options for the builder to process.

        Returns:
            Instance of the named builder.
        """

        class_name = 'rpmlb.builder.{0}.{1}Builder'.format(
            name,
            utils.camelize(name)
        )
        instance = utils.get_instance(class_name, work, **options)

        LOG.debug('Loaded builder with %s', name)
        return instance

    def run(self, work, **kwargs):
        is_resume = kwargs.get('resume', False)

        if is_resume:
            message = (
                'Skip the process before build, '
                'because the resume option was used.'
            )
            LOG.info(message)
        else:
            self.before(work, **kwargs)

        for package_dict, num_name in work.each_package_dir():
            try:
                if self._is_build_skipped(package_dict, num_name,
                                          is_resume, **kwargs):
                    continue

                self.prepare(package_dict, **kwargs)

                # Repeat the call in case of failure
                retry_count = kwargs.get('retry', 0)
                retry_call(
                    self.build, fargs=(package_dict,), fkwargs=kwargs,
                    tries=retry_count+1,  # 0 tries == no call
                )

            except Exception:
                message = 'pacakge_dict: {0}, num: {1}, work_dir: {2}'.format(
                    package_dict, num_name, work.working_dir)
                error = RuntimeError(message)
                tb = sys.exc_info()[2]
                error = error.with_traceback(tb)
                raise error

        self.after(work, **kwargs)
        return True

    def before(self, work: Work, **options):
        """One-time build setup.

        This method is intended for setting up the whole build process,
        and it is skipped when using the resume.
        For preparation/initialization of the builder itself, use `__init__`.

        Keyword arguments:
            work: Overview of the work to do.
            options: Command line option for the builder to process.
        """

    def after(self, work: Work, **options):
        """Final build clean-up.

        This method is intended for cleaning up after successful build,
        and it is called once all the work is completed.

        Keyword arguments:
            work: Overview of the work that has been done.
            options: Command line option for the builder to process.
        """

    def prepare(self, package_dict: Mapping[str, Any], **options):
        """Prepare single package for a build.

        Keyword arguments:
            package_dict: A dictionary of package metadata.
            options: Command line option for the builder to process.
        """

        if 'name' not in package_dict:
            raise ValueError('package_dict is invalid.')

        spec_file_path = Path('{name}.spec'.format_map(package_dict))

        with self.edit_spec_file(spec_file_path) as (source_file, target_file):
            content_stream = iter(source_file)  # Start modifications

            if 'replaced_macros' in package_dict:
                content_stream = self.replace_macros(
                    content_stream,
                    package_dict['replaced_macros'],
                )

            if 'macros' in package_dict:
                content_stream = self.add_macros(
                    content_stream,
                    package_dict['macros'],
                )

            # Perform any extra edits needed by derived builder
            content_stream = self.prepare_extra_steps(
                content_stream,
                package_dict,
            )

            target_file.write(''.join(content_stream))  # End modifications

        if 'cmd' in package_dict:
            env = {}
            dist = options.get('dist')
            if dist:
                env['DIST'] = dist
            Yaml.run_cmd_element(package_dict['cmd'], env=env)

    def build(self, package_dict, **kwargs):
        raise NotImplementedError('Implement this method.')

    @staticmethod
    @contextmanager
    def edit_spec_file(target_path: Path):
        """Safely edit a SPEC file in-place.

        The target is backed up as '{target}.orig' if needed.

        Keyword arguments:
            target_path: The modified SPEC file path.

        Returns:
            Context manager providing open handles
            for input and output file.
        """

        # Ensure path type
        if not isinstance(target_path, Path):
            target_path = Path(target_path)

        # Back up the original
        source_path = target_path.with_suffix('.spec.orig')
        if not source_path.exists():
            target_path.rename(source_path)

        # Provide the handles
        with source_path.open(mode='r') as source_file, \
                target_path.open(mode='w') as target_file:
            print('# Edited by rpmlb', file=target_file)

            yield source_file, target_file

            # Ensure that all unprocessed source contents
            # are written to the target
            target_file.write(source_file.read())

    @staticmethod
    def add_macros(source: Iterator[str], macros: Mapping[str, str]):
        """Add macro definitions to the source stream.

        Keyword arguments:
            source: The source file iterator.
            macros: Mapping of macro name to macro definition
                    for all macros to add.

        Yields:
            Lines of the modified file.
        """

        # Prepend all definitions before the stream
        for name, value in macros.items():
            yield '%global {name} {value}\n'.format(name=name, value=value)

        # Pass the rest of the file
        yield from source

    @staticmethod
    def replace_macros(source: Iterator[str], macros: Mapping[str, str]):
        """Replace macros in the source stream with new values.

        Keyword arguments:
            source: The source file iterator.
            macros: Mapping of macro name to macro definition
                    for all macros to be replaced.

        Yields:
            Lines of the modified file.
        """

        def replacement(match: Match) -> str:
            """Macro replacement logic"""

            macro_name = match.group('name')
            macro_value = macros.get(macro_name, match.group('value'))
            return '%global {} {}'.format(macro_name, macro_value)

        # Need whole file contents for matching multi-line macros
        contents = ''.join(source)

        # Substitute all macros
        contents = MACRO_REGEX.sub(replacement, contents)

        # Pass the modified lines
        yield from contents.splitlines(keepends=True)

    def prepare_extra_steps(
        self,
        source: Iterator[str],
        package_metadata: Mapping[str, Any]
    ):
        """Builder-specific package preparation.

        Override if needed.

        Keyword arguments:
            source: The source file iterator.
            package_metadata: The metadata of the prepared package,
                as passed to prepare().
        """

        yield from source  # pass for generators

    def _is_build_skipped(self, package_dict: dict, num_name: str,
                          is_resume: bool, **kwargs):
        """Return if skip a build for a pacakge.

        Override if needed.

        Keyword arguments:
            package_dict: A dictionary of package metadata.
            num_name: A package's order number string.
            is_redume: if redume or not.
            kwargs: option arguments.
        """

        if not package_dict:
            raise ValueError('package_dict is required.')
        if not num_name:
            raise ValueError('num_name is required.')

        resume_num = 0
        if is_resume:
            resume_num = kwargs['resume']
        dist = kwargs.get('dist')

        is_skipped = False
        if is_resume:
            num = int(num_name)
            if num < resume_num:
                is_skipped = True
        elif dist and 'dist' in package_dict:
            pattern = re.compile(package_dict['dist'])
            if not pattern.match(dist):
                is_skipped = True
        return is_skipped



import logging

from .. import utils
from .base import BaseDownloader

LOG = logging.getLogger(__name__)


class BaseRpkgDownloader(BaseDownloader):
    """A downloader class to get a pacakge with package command."""

    @property
    def command(self):
        raise NotImplementedError('Implement this method.')

    def download(self, package_dict, **kwargs):
        if not package_dict:
            raise ValueError('package_dict is required.')
        if 'branch' not in kwargs or not kwargs['branch']:
            raise ValueError('branch is required.')
        branch = kwargs['branch']

        package = package_dict['name']
        cmd = r'''
{} co {} && \
cd {} && \
git checkout {}
        '''.strip().format(self.command, package, package, branch)
        utils.run_cmd(cmd)



import logging
import os.path

from rpmlb.yaml import Yaml

LOG = logging.getLogger(__name__)


class Custom:
    """A class to manage custom file."""

    def __init__(self, file_path):
        self._file_path = file_path = os.path.abspath(str(file_path))
        self._custom_dir = os.path.dirname(file_path)
        self._yaml_content = None

    def run_cmds(self, key, **kwargs):
        env = {}
        # Support environment variable to use PKG as pacakge name in the file.
        if 'name' in kwargs:
            env['PKG'] = kwargs['name']

        cmd_dict = self.yaml_content
        if key not in cmd_dict:
            return
        cmd_element = cmd_dict[key]

        env['CUSTOM_DIR'] = self._custom_dir
        Yaml.run_cmd_element(cmd_element, env=env)

    @property
    def yaml_content(self):
        if self._yaml_content:
            return self._yaml_content
        yaml = Yaml(self._file_path)
        self._yaml_content = yaml.content
        return self._yaml_content



"""CLI interface for the package"""

import logging
import os
import re

import click

from . import logging as rpmlb_logging
from .builder.base import BaseBuilder
from .downloader.base import BaseDownloader
from .recipe import Recipe
from .work import Work

DIST_RE = re.compile('^(fc|el|centos)[0-9]*$')


@click.command(add_help_option=False)
# General options
@click.help_option('--help', '-h')  # Enable short help switch (-h)
@click.version_option()
@click.option(
    '--verbose', '-v', is_flag=True, default=False,
    help='Turn on verbose logging.',
    # Enable logging as early as possible
    is_eager=True, expose_value=False,
    callback=lambda ctx, param, verbose:
        rpmlb_logging.configure_logging(verbose),
)
@click.option(
    '--download', '-d',
    type=click.Choice('none local fedpkg rhpkg custom'.split()),
    default='none',
    help='Choose a download type.',
)
@click.option(
    '--build', '-b',
    type=click.Choice('dummy mock copr koji custom'.split()),
    default='dummy',
    help='Choose a build type.',
)
@click.option(
    '--work-directory', '-w', metavar='WORK_DIRECTORY',
    type=click.Path(exists=True, file_okay=False, writable=True,
                    resolve_path=True),
    default=None,
    help='Specify a working directory.',
)
@click.option(
    '--custom-file', '-c', metavar='CUSTOM_FILE',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help='Instructions for custom downloader and builder.',
)
@click.option(
    '--dist', metavar=DIST_RE.pattern,
    help='Specifiy a distribution platform considered in a recipe file.',
    callback=lambda ctx, param, dist: validate_dist(dist),
)
# Download options
@click.option(
    '--branch', '-B', metavar='BRANCH',
    help=('Git branch for downloaders that use it '
          '(download type: fedpkg, rhpkg).'),
)
@click.option(
    '--source-directory', '-S', metavar='SOURCE_DIRECTORY',
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    default=os.path.abspath(os.getcwd()),
    help='Package source directory for local downloader.',
)
# Build options
@click.option(
    '--resume', '-r', metavar='RESUME',
    type=click.INT,
    help='Resume build from specified position.',
)
@click.option(
    '--retry', metavar='RETRY_COUNT',
    type=click.IntRange(min=0), default=0,  # no negatives
    help='Retry failed build RETRY_COUNT times before aborting.',
)
@click.option(
    '--pkg-cmd',
    type=click.Choice('fedpkg rhpkg'.split()),
    help='Choose a package command.',
)
@click.option(
    '--mock-config', '-M', metavar='MOCK_CONFIG',
    help='Mock configuration for mock builder.',
)
@click.option(
    '--copr-repo', '-C', metavar='COPR_REPO',
    help='Target Copr for copr builder.',
)
@click.option(
    '--koji-epel', metavar='VERSION', type=int, default=7,
    help='Target EPEL version for koji builder.',
)
@click.option(
    '--koji-profile', metavar='PROFILE',
    help='Configuration profile for koji builder.',
)
@click.option(
    '--koji-target-format', metavar='FORMAT',
    help='Target format string for koji builder.',
)
@click.option(
    '--koji-scratch', is_flag=True, default=False,
    help='Make scratch builds with koji builder.',
)
@click.option(
    '--koji-owner', metavar='OWNER',
    help='Owner of packages added by koji builder.',
)
# Positional arguments
@click.argument(
    'recipe_file',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)
@click.argument('collection_id')
def run(recipe_file, collection_id, **option_dict):
    """Download and build RPMs listed in RECIPE_FILE under COLLECTION_ID
    (such as 'python33').
    """

    log = logging.getLogger(__name__)

    _choose_pkg_cmd(option_dict)

    # Load recipe and processing objects
    recipe = Recipe(recipe_file, collection_id)
    recipe.verify()

    # Prepare the working directory
    # HINT: with contextlib.closing(Work(recipe, **option_dict)) as work:
    work = Work(recipe, **option_dict)

    builder = BaseBuilder.get_instance(
        option_dict['build'],
        work,
        **option_dict
    )
    downloader = BaseDownloader.get_instance(option_dict['download'])

    # Download
    log.info('Downloading...')
    downloader.run(work, **option_dict)

    # Build
    log.info('Building...')
    builder.run(work, **option_dict)

    log.info('Success!')


def validate_dist(dist):
    if dist and not DIST_RE.match(dist):
        raise click.BadParameter(
            'dist shoule be match to pattern {0}'.format(DIST_RE.pattern))
    return dist


def _choose_pkg_cmd(option_dict):
    if not option_dict:
        return
    # User chose the command explicitly.
    if 'pkg_cmd' in option_dict and option_dict['pkg_cmd']:
        return

    # Set default value.
    option_dict['pkg_cmd'] = 'fedpkg'

    download = option_dict.get('download', None)
    # No explicit command, but relevant builder was selected.
    if download in ('fedpkg', 'rhpkg'):
        option_dict['pkg_cmd'] = download



import logging

from .. import utils
from ..work import Work
from .base import BaseBuilder

LOG = logging.getLogger(__name__)


class MockBuilder(BaseBuilder):
    """A builder class for Mock."""

    def __init__(self, work: Work, mock_config: str = None, **options):
        """Initialize the builder.

        Keyword arguments:
            work: The overview of the work to do.
            mock_config: Name of the Mock configuration profile to use.
        """

        super().__init__(work, mock_config=mock_config, **options)

        if mock_config is None:
            raise ValueError('mock_config is required.')

        #: Name of the Mock configuration profile to use
        self.mock_config = mock_config

    def before(self, work, **kwargs):
        utils.run_cmd('mock -r {} --scrub=all'.format(self.mock_config))

    def build(self, package_dict, **kwargs):
        utils.run_cmd('rm -v *.rpm', check=False)
        utils.run_cmd('{} srpm'.format(self._pkg_cmd))
        utils.run_cmd('mock -r {} -n *.rpm'.format(self.mock_config))



from unittest import mock

import pytest

from rpmlb.custom import Custom


@pytest.fixture
def valid_custom(valid_custom_file_path):
    """A Custom object with a valid custom file."""
    return Custom(valid_custom_file_path)


def test_init_loads_file(valid_custom):
    assert valid_custom


def test_run_cmds_runs_cmds_on_valid_custom_file(valid_custom, monkeypatch):
    with pytest.helpers.generate_tmp_path_list(2) as tmp_paths:
        random_file_foo, random_file_bar_part = tmp_paths

        content = {
            'build': [
                'touch {0!s}'.format(random_file_foo),
                'touch "{0!s}-$PKG"'.format(random_file_bar_part),
            ]
        }
        monkeypatch.setattr(type(valid_custom), 'yaml_content',
                            mock.PropertyMock(return_value=content))

        valid_custom.run_cmds('build', name='rubygem-bar')

        random_file_bar = random_file_bar_part.with_name(
            random_file_bar_part.name + '-rubygem-bar'
        )

        assert random_file_foo.is_file()
        assert random_file_bar.is_file()


def test_run_cmds_skips_cmds_on_unknown_key(valid_custom, random_file_path,
                                            monkeypatch):
    content = {
        'build': [
            'touch {0!s}'.format(random_file_path),
        ]
    }
    monkeypatch.setattr(type(valid_custom), 'yaml_content',
                        mock.PropertyMock(return_value=content))

    valid_custom.run_cmds('dummy')

    assert not random_file_path.is_file()


def test_yaml_content_returns_content(valid_custom):
    content = valid_custom.yaml_content
    assert content


def test_yaml_content_returns_singleton_object(valid_custom):
    content = valid_custom.yaml_content
    content2 = valid_custom.yaml_content
    assert content is content2



"""Tests for the KojiBuilder"""

import shutil
from collections import namedtuple
from itertools import zip_longest
from logging import DEBUG
from pathlib import Path
from textwrap import dedent

import pytest

from rpmlb.builder.koji import KojiBuilder
from rpmlb.logging import configure_logging
from rpmlb.recipe import Recipe
from rpmlb.work import Work


@pytest.fixture(
    params=[
        ('tests/fixtures/specfiles/minimal.spec', 'test.spec'),
        ('tests/fixtures/specfiles/metapackage.spec', 'rh-ror50.spec'),
    ],
    ids=[
        'minimal',
        'metapackage'
    ],
)
def spec_path(request, tmpdir):
    """Provide a SPEC file path according to parameter."""

    # Copy the SPEc to temporary directory
    source = Path(request.param[0])
    target = Path(str(tmpdir), request.param[1])

    shutil.copyfile(str(source), str(target))  # no support for pathlib :(

    with tmpdir.as_cwd():
        yield target


@pytest.fixture
def spec_contents(spec_path):
    """Provide text contents of a minimal spec file"""

    with spec_path.open() as istream:
        return istream.read()


@pytest.fixture
def collection_id():
    """Testing collection identification"""

    return 'rh-ror50'


@pytest.fixture
def work(valid_recipe_path, collection_id):
    """Provide Work instance."""

    valid_recipe = Recipe(str(valid_recipe_path), collection_id)
    return Work(valid_recipe)


@pytest.fixture
def builder(work):
    """Provide minimal KojiBuilder instance."""

    return KojiBuilder(work, koji_epel=7, koji_owner='test-owner')


@pytest.fixture(params=[
    KojiBuilder.DEFAULT_TARGET_FORMAT,
    'test-target',
    'fedora-{collection}-epel{epel}',
])
def target_format(request):
    """Koji target format"""

    return request.param


@pytest.fixture(params=list(range(6, 8)))
def epel(request):
    """EPEL version number"""

    return request.param


@pytest.fixture(params=[None, 'koji', 'cbs'])
def profile(request):
    """Koji profile name"""

    return request.param


@pytest.fixture(params=[True, False])
def scratch_build(request):
    """Scratch build indicator"""

    return request.param


@pytest.fixture
def cli_parameters(target_format, epel, profile, scratch_build):
    """Builder parameters as coming from the CLI"""

    return {
        'koji_epel': epel,
        'koji_owner': 'test-owner',
        'koji_profile': profile,
        'koji_target_format': target_format,
        'koji_scratch': scratch_build,

        # Extra parameter to assure the builder can take them
        'copr_repo': 'test/example',
    }


def test_init_sets_attributes(work, cli_parameters):
    """Ensure that the parameters are set to appropriate values."""

    builder = KojiBuilder(work, **cli_parameters)

    assert builder.collection == work._recipe._collection_id
    assert builder.epel == cli_parameters['koji_epel']
    assert builder.owner == cli_parameters['koji_owner']
    assert builder.profile == cli_parameters['koji_profile']
    assert builder.target_format == cli_parameters['koji_target_format']
    assert builder.scratch_build == cli_parameters['koji_scratch']


@pytest.mark.parametrize('required', ['koji_epel', 'koji_owner'])
def test_init_checks_parameters(required, work, cli_parameters):
    """Ensure that required parameters are checked by __init__."""

    del cli_parameters[required]

    with pytest.raises(ValueError):
        KojiBuilder(work, **cli_parameters)


@pytest.mark.parametrize('profile,expected_command', [
    (None, ['koji']),
    ('koji', ['koji', '--profile', 'koji']),
    ('cbs', ['koji', '--profile', 'cbs']),
])
def test_base_command_respect_profile(builder, profile, expected_command):
    """Ensure that the base_command attribute respects profile attribute."""

    builder.profile = profile
    assert builder.base_command == expected_command


def test_default_format_is_used(work, epel):
    """Default format is used when no target format is specified."""

    parameters = {
        'koji_epel': epel, 'koji_owner': 'test-owner',  # required
        'koji_target_format': None,  # explicit None => should use default
    }

    builder = KojiBuilder(work, **parameters)
    assert builder.target_format == KojiBuilder.DEFAULT_TARGET_FORMAT


@pytest.mark.parametrize('collection', ['test', 'rh-ror50'])
def test_target_respects_format(
    builder, target_format, collection, epel
):
    """Ensure that the target is properly constructed from parameters."""

    builder.target_format = target_format
    builder.collection = collection
    builder.epel = epel

    expected_target = target_format.format(collection=collection, epel=epel)
    assert builder.target == expected_target


def test_make_srpm_creates_srpm(spec_path, collection_id, epel):
    """Ensure that make_srpm works as expected"""

    configure_logging(DEBUG)

    arguments = {
        'name': spec_path.stem,
        'collection': collection_id,
        'epel': epel,
    }

    possible_names = {
        # Metapackage
        '{collection}-1-1.el{epel}.src.rpm'.format_map(arguments),
        # Regular package
        '{collection}-{name}-1.0-1.el{epel}.src.rpm'.format_map(arguments),
    }

    srpm_path = KojiBuilder._make_srpm(**arguments)

    assert srpm_path.exists(), srpm_path
    assert srpm_path.name in possible_names


def test_missing_spec_is_reported(tmpdir):
    """make_srpm does not attempt to build nonexistent SPEC file"""

    configure_logging(DEBUG)

    with tmpdir.as_cwd(), pytest.raises(FileNotFoundError):
        KojiBuilder._make_srpm(name='test', collection='test', epel=7)


def test_prepare_adjusts_bootstrap_release(builder, spec_contents):
    """Release is adjusted on bootstrap build."""

    lines = spec_contents.splitlines(keepends=True)
    lines = builder.prepare_extra_steps(lines, {
        'name': 'test',
        'bootstrap_position': 1,
    })

    release_line, = filter(lambda l: l.startswith('Release:'), lines)

    assert '.bs01%{?dist}' in release_line


@pytest.mark.parametrize('scratch,expected_commands', [
    (True, ['koji add-pkg', 'koji build']),
    (False, ['koji add-pkg', 'koji build', 'koji wait-repo']),
])
def test_build_emit_correct_commands(
    monkeypatch, builder, scratch, expected_commands
):
    """Builder emits expected commands on build."""

    # Gather all emitted commands
    commands = []
    monkeypatch.setattr(
        'rpmlb.builder.koji.run_cmd',
        lambda cmd, **__: commands.append(cmd),
    )

    # Skip make_srpm, as it requires the run_cmd to actually do something
    monkeypatch.setattr(
        KojiBuilder, '_make_srpm',
        lambda *_, **__: Path('test.src.rpm'),
    )

    # Provide hardcoded destination tag
    monkeypatch.setattr(KojiBuilder, '_destination_tag', 'test_tag_name')

    builder.scratch_build = scratch
    builder.build({'name': 'test'})

    command_pairs = zip_longest(commands, expected_commands)
    assert all(cmd.startswith(exp) for cmd, exp in command_pairs), commands


def test_destination_tag_parsed_correctly(monkeypatch, builder):
    """Assert that the destination tag is correctly extracted from output."""

    # Simulate output of `koji list-targets`
    raw_output = namedtuple('MockCommandOutput', ['stdout', 'stderr'])(
        stderr=b'',
        stdout=dedent('''\
        Name                Buildroot           Destination
        ----------------------------------------------------------
        test-target         test-target-build   test-tag-candidate
        ''').encode('utf-8')
    )

    monkeypatch.setattr(
        'rpmlb.builder.koji.run_cmd_with_capture',
        lambda *_, **__: raw_output,
    )

    builder.target_format = 'test-target'
    assert builder._destination_tag == 'test-tag-candidate'


def test_add_package_respects_owner(monkeypatch, builder):
    """Assert that the owner is used by the _add_package method."""

    commands = []
    monkeypatch.setattr(
        'rpmlb.builder.koji.run_cmd',
        lambda cmd, **__: commands.append(cmd),
    )
    monkeypatch.setattr(
        KojiBuilder, '_destination_tag', 'test-destination'
    )

    builder.owner = 'expected-owner'
    builder._add_package(name='test')

    assert len(commands) == 1

    cmd, = commands
    assert '--owner' in cmd
    assert builder.owner in cmd



import logging
from collections import Counter
from itertools import starmap
from typing import Iterator, Mapping, Union

import click

from rpmlb.yaml import Yaml

LOG = logging.getLogger(__name__)


class Recipe:
    """A class to describe recipe data."""

    ERROR_REQUIRED = '{} is required.'
    ERROR_EMPTY = '{} is empty.'
    ERROR_TYPE = '{} should be a {}.'
    ERROR_INVALID = '{} is invalid.'
    ERROR_UNKNOWN = '{} is an unknown element.'
    TYPE_DISP_MAP = {
        'str': 'string',
        'list': 'sequences',
        'dict': 'mappings',
    }

    def __init__(self, file_path, collection_id):
        if not file_path:
            raise ValueError('file_path is required.')
        if not collection_id:
            raise ValueError('collection_id is required.')

        self._collection_id = collection_id

        yaml = Yaml(file_path)
        LOG.debug('Loaded recipe: %s', file_path)
        recipe_dict = yaml.content
        self.recipe = recipe_dict[collection_id]
        self.num_of_package = len(self.recipe['packages'])

    def each_normalized_package(self):
        """Present recipe packages in normalized form.

        Yields:
            Dictionary with the package's data.

        Common keys in the yielded dictionary:
            - name: The package's name
            - bootstrap_position: Numerical position in the bootstrapping
                sequence, or None if no bootstrap is necessary.
            - macros, replaced_macros: Mapping of macro name to (new)
                macro body.
        """

        packages = self.recipe['packages']
        bootstrap_map = self._count_bootstrap_sequences()

        for package in packages:
            name = self._package_name(package)
            bootstrap_position = next(bootstrap_map[name], None)

            package_dict = {
                'name': name,
                'bootstrap_position': bootstrap_position,
            }

            if isinstance(package, Mapping):
                # package should be one-item dictionary
                metadata, = package.values()
                package_dict.update(metadata)

            yield package_dict

    def verify(self):
        error_messages = self._verify_recipe()
        if error_messages:
            raise RecipeError(error_messages)

        return True

    def _verify_recipe(self) -> list:
        error_messages = []
        recipe = self.recipe

        if 'packages' not in recipe:
            message = self._format_error_required('packages')
            error_messages.append(message)

        for key in recipe:
            is_known_key = False
            if key == 'name':
                is_known_key = True
                if not recipe[key]:
                    message = self._format_error_empty(key)
                    error_messages.append(message)
                if not isinstance(recipe[key], str):
                    message = self._format_error_type(key, str)
                    error_messages.append(message)
            if key == 'requires':
                is_known_key = True
                if not recipe[key]:
                    message = self._format_error_empty(key)
                    error_messages.append(message)
                if not isinstance(recipe[key], list):
                    message = self._format_error_type(key, list)
                    error_messages.append(message)
            if key == 'packages':
                is_known_key = True
                is_packages_valid = True
                if not recipe[key]:
                    is_packages_valid = False
                    message = self._format_error_empty(key)
                    error_messages.append(message)
                if not isinstance(recipe[key], list):
                    is_packages_valid = False
                    message = self._format_error_type(key, list)
                    error_messages.append(message)
                if is_packages_valid:
                    # Check the detail of packages element.
                    error_messages = self._verify_packages_and_append(
                        error_messages, recipe[key]
                    )
            if not is_known_key:
                message = self._format_error_unknown(key)
                error_messages.append(message)
        return error_messages

    def _verify_packages_and_append(self, messages: list,
                                    packages: list) -> list:
        if messages is None:
            raise ValueError('messages is required.')
        if not packages:
            raise ValueError('packages is required.')
        error_messages = messages.copy()
        disp_packages_format = 'packages[{}] {}'

        for index, package in enumerate(packages):
            disp_num = index + 1

            if isinstance(package, str):
                disp_package = disp_packages_format.format(disp_num, package)
                if not package:
                    message = self._format_error_empty(disp_package)
                    error_messages.append(message)
            elif isinstance(package, dict):
                name = list(package.keys())[0]
                disp_package = disp_packages_format.format(disp_num, name)

                if len(package) != 1:
                    message = self._format_error_invalid(disp_package)
                    error_messages.append(message)
                else:
                    package_metadata = package[name]
                    error_messages = self._verify_packages_metadata_and_append(
                        error_messages, package_metadata, disp_package,
                    )
            else:
                disp_package = disp_packages_format.format(disp_num, package)

                message = self._format_error_type(disp_package, (str, dict))
                error_messages.append(message)
        return error_messages

    def _verify_packages_metadata_and_append(self, messages: list,
                                             metadata: dict,
                                             disp_package: str) -> list:
        if messages is None:
            raise ValueError('messages is required.')
        if not metadata:
            raise ValueError('metadata is required.')
        if not disp_package:
            raise ValueError('disp_package is required.')
        error_messages = messages.copy()

        for key in metadata.keys():
            disp_name = '{} {}'.format(disp_package, key)
            is_known_key = False
            if key in {'macros', 'replaced_macros'}:
                is_known_key = True
                if not metadata[key]:
                    message = self._format_error_empty(disp_name)
                    error_messages.append(message)
                if not isinstance(metadata[key], dict):
                    message = self._format_error_type(disp_name, dict)
                    error_messages.append(message)
            if key == 'cmd':
                is_known_key = True
                if not metadata[key]:
                    message = self._format_error_empty(disp_name)
                    error_messages.append(message)
                if not isinstance(metadata[key], (str, list)):
                    message = self._format_error_type(disp_name, (str, list))
                    error_messages.append(message)
            if key == 'dist':
                is_known_key = True
                if not metadata[key]:
                    message = self._format_error_empty(disp_name)
                    error_messages.append(message)
                if not isinstance(metadata[key], str):
                    message = self._format_error_type(disp_name, str)
                    error_messages.append(message)
            if not is_known_key:
                message = self._format_error_unknown(disp_name)
                error_messages.append(message)

        return error_messages

    def _format_error_required(self, name):
        return self.ERROR_REQUIRED.format(name)

    def _format_error_empty(self, name):
        return self.ERROR_EMPTY.format(name)

    def _format_error_type(self, name, types: Union[type, tuple]):
        disp_type = None
        if type(types) == type:
            type_classes = tuple([types])
        else:
            type_classes = types

        def make_type_disp(cls):
            return self.TYPE_DISP_MAP[cls.__name__]

        class_strs = list(map(make_type_disp, type_classes))
        disp_type = ' or '.join(class_strs)

        return self.ERROR_TYPE.format(name, disp_type)

    def _format_error_invalid(self, name):
        return self.ERROR_INVALID.format(name)

    def _format_error_unknown(self, name):
        return self.ERROR_UNKNOWN.format(name)

    def _package_name(self, package: Union[str, dict]) -> str:
        """Extract package name.

        Keyword arguments:
            package: Raw package data to analyze.
                It should be either string or one-item dictionary.

        Returns:
            Name of the package.

        Raises:
            ValueError: Invalid package type.
        """

        if isinstance(package, str):  # The name itself
            return package

        elif isinstance(package, dict):
            # package should be one-item dictionary
            if len(package) != 1:
                message = self._format_error_invalid(package)
                raise ValueError(message)

            # The name is the only key
            name, = package.keys()
            return name

        else:
            message = self._format_error_type(package, (str, dict))
            raise ValueError(message)

    def _count_bootstrap_sequences(self) -> Mapping[str, Iterator[int]]:
        """Count package instances and produce bootstrap sequences.

        Returns:
            Mapping of package number to iterator of bootstrap numbers.
        """

        name_list = map(self._package_name, self.recipe['packages'])
        count_map = Counter(name_list)

        def make_sequence(name: str, count: int):
            """Count package instances except the last one; start from 1"""

            sequence = range(1, count)
            return name, iter(sequence)

        # Create a dictionary from (name, sequence) tuples
        sequence_map = dict(starmap(make_sequence, count_map.items()))

        return sequence_map


class RecipeError(click.ClickException):
    """A class to manage validation error for recipe data."""

    def __init__(self, messages: list):
        super().__init__(messages)
        self.messages = messages

    def format_message(self):
        return '\n'.join(['Recipe file is invalid.'] + self.messages)



from unittest import mock

import pytest

from rpmlb.builder.mock import MockBuilder
from rpmlb.recipe import Recipe
from rpmlb.work import Work


@pytest.fixture
def work(valid_recipe_path):
    """Provide Work instance."""

    valid_recipe = Recipe(str(valid_recipe_path), 'rh-ror50')
    return Work(valid_recipe)


@pytest.fixture
def builder(work):
    pkg_cmd = 'fedpkg'
    mock_config = 'fedora-rawhide-x86_64'
    return MockBuilder(work, pkg_cmd=pkg_cmd, mock_config=mock_config)


def test_init(builder):
    assert builder


def test_init_raises_error_without_mock_config(builder):
    pkg_cmd = 'fedpkg'
    with pytest.raises(ValueError):
        MockBuilder(work, pkg_cmd=pkg_cmd)


def test_before_passes(builder, work):
    with mock.patch('rpmlb.utils.run_cmd', mock.Mock(return_value=True)):
        builder.before(work)
    assert True


def test_build_passes(builder, work):
    with mock.patch('rpmlb.utils.run_cmd',
                    mock.Mock(return_value=True)) as mock_run_cmd:
        package_dict = {'name': 'a'}
        builder.build(package_dict)
        mock_run_cmd.assert_any_call('rm -v *.rpm', check=False)
        mock_run_cmd.assert_any_call('fedpkg srpm')
        mock_run_cmd.assert_any_call('mock -r fedora-rawhide-x86_64 -n *.rpm')

    assert True



"""Test argument parsing"""

import logging
import os
from contextlib import ExitStack
from pathlib import Path
from textwrap import dedent

import click
import pytest
from click.testing import CliRunner

from rpmlb import cli


@pytest.fixture
def root_logger():
    return logging.getLogger()


@pytest.fixture
def runner(root_logger):
    r = CliRunner()

    with r.isolated_filesystem():
        yield r


@pytest.fixture
def recipe_path(tmpdir):
    """Test recipe file path"""

    with tmpdir.as_cwd():
        path = Path.cwd().resolve() / 'recipe.yml'

        # Fill in valid but empty recipe content
        with path.open(mode='w', encoding='utf-8') as outfile:
            contents = dedent('''\
                test:
                    name: Test recipe
                    requires: []
                    packages:
                        - test
                ''')
            print(contents, file=outfile)

        yield path


@pytest.fixture
def recipe_arguments(recipe_path):
    """Common recipe arguments: Path to the recipe file and recipe name."""

    return [str(recipe_path), 'test']


@pytest.mark.parametrize('option', ('recipe_file', 'collection_id', 'build',
                                    'download', 'branch', 'source_directory',
                                    'pkg_cmd'))
def test_parse_argv_no_options(tmpdir, option):
    """Tests proper default values of the CLI"""

    recipe_file = tmpdir.join('ror.yml')
    collection_id = 'rh-ror50'

    # Prepare environment
    recipe_file.write('')

    current_dir = os.path.abspath(os.getcwd())

    expected = {
        'recipe_file': str(recipe_file),
        'collection_id': str(collection_id),
        'build': 'dummy',
        'download': 'none',
        'branch': None,
        'source_directory': current_dir,
        'pkg_cmd': None,
    }

    # Parse the arguments
    with tmpdir.as_cwd():
        argv = list(map(str, [recipe_file, collection_id]))
        args = cli.run.make_context('rpmlb', argv).params

        assert args[option] == expected[option]


@pytest.mark.parametrize('verbose', (True, False))
def test_log_verbosity(root_logger, runner, verbose, recipe_arguments):
    """Ensure that the verbosity is set properly on."""

    verbose_args = ['--verbose'] if verbose else []
    level = logging.DEBUG if verbose else logging.INFO

    runner.invoke(cli.run, verbose_args + recipe_arguments)
    assert root_logger.getEffectiveLevel() == level


@pytest.fixture(params=('work-directory', 'custom-file'))
def path_kind(request):
    return request.param


def path_options(path_kind):
    """Excepted environment for the path tests"""
    filename = 'custom.yml' if path_kind == 'custom-file' else path_kind

    root = Path.cwd().resolve()
    path = root/filename

    options = ['--' + path_kind, str(path)]

    assert not path.exists()

    return path, options


def test_path_nonexistent(runner, path_kind, recipe_arguments):
    path, options = path_options(path_kind)

    result = runner.invoke(cli.run, options + recipe_arguments,
                           standalone_mode=False)
    assert result.exception is not None
    assert isinstance(result.exception, click.BadParameter)


def test_path_expected_and_absolute(path_kind, recipe_arguments):
    path, options = path_options(path_kind)

    if path_kind.endswith('directory'):
        path.mkdir()
    else:
        path.touch(mode=0o600)

    with cli.run.make_context('test-{}-ok'.format(path_kind),
                              options + recipe_arguments) as ctx:
        result = Path(ctx.params[path_kind.replace('-', '_')])
    assert result == path
    assert result.is_absolute()


def test_path_bad_permissions(runner, path_kind, recipe_arguments):
    path, options = path_options(path_kind)

    if path_kind.endswith('directory'):
        # unwritable directory
        path.mkdir()
        path.chmod(0o500)
    else:
        # unreadable file
        path.touch(mode=0o200)

    result = runner.invoke(cli.run, options + recipe_arguments,
                           standalone_mode=False)
    assert result.exception is not None
    assert isinstance(result.exception, click.BadParameter)


@pytest.mark.parametrize('pkg_cmd', ('fedpkg', 'rhpkg'))
def test_pkg_cmd_returns_valid_value(recipe_arguments, pkg_cmd):
    options = ['--pkg-cmd', pkg_cmd]
    with cli.run.make_context('test_pkg_cmd_{}_returns_value'.format(pkg_cmd),
                              options + recipe_arguments) as ctx:
        assert ctx.params['pkg_cmd'] == pkg_cmd


def test_pkg_cmd_raises_error_on_invalid_options(runner, recipe_arguments):
    options = ['--pkg-cmd', 'foo']
    result = runner.invoke(cli.run, options + recipe_arguments,
                           standalone_mode=False)
    assert result.exception is not None
    assert isinstance(result.exception, click.BadParameter)


@pytest.mark.parametrize('valid_dist',
                         ('fc', 'fc26', 'el', 'el7', 'centos', 'centos7'))
def test_dist_returns_value_on_valid_options(recipe_arguments, valid_dist):
    options = ['--dist', valid_dist]
    with cli.run.make_context('test_dist_returns_value_on_valid_options',
                              options + recipe_arguments) as ctx:
        assert isinstance(ctx.params['dist'], str)


def test_dist_raises_error_on_invalid_value(runner, recipe_arguments):
    options = ['--dist', 'fedora']
    result = runner.invoke(cli.run, options + recipe_arguments,
                           standalone_mode=False)
    assert result.exception is not None
    assert isinstance(result.exception, click.BadParameter)


def test_resume_conversion(recipe_arguments):
    """Resume is converted into integer value."""

    options = ['--resume', '42']
    with cli.run.make_context('test-resume-conversion',
                              options + recipe_arguments) as ctx:
        assert isinstance(ctx.params['resume'], int)


def test_invalid_resume(runner, recipe_arguments):

    options = ['--resume', 'start']

    result = runner.invoke(cli.run, options + recipe_arguments,
                           standalone_mode=False)
    assert result.exception is not None
    assert isinstance(result.exception, click.BadParameter)


@pytest.mark.parametrize('option,value', [
    ('branch', 'sclo7-rh-nodejs4-el7'),
    ('mock-config', 'default'),
    ('copr-repo', 'scratch-ror5'),
])
def test_simple_options(recipe_arguments, option, value):
    """Specific option values are passed unprocessed."""

    options = ['--' + option, value]
    with cli.run.make_context('test-{}-passing'.format(option),
                              options + recipe_arguments) as ctx:

        assert ctx.params[option.replace('-', '_')] == value


@pytest.mark.parametrize('help_switch', ['--help', '-h'])
def test_help_option_variants(runner, help_switch, recipe_arguments):
    """Test that both short and long version of help switch works."""

    # NOTE: recipe_arguments are necessary, both run.make_context and
    # runner.invoke break without them for some reason
    result = runner.invoke(cli.run, [help_switch] + recipe_arguments)

    assert result.exit_code == 0, result.output


@pytest.mark.parametrize('version_switch', ['--version'])
def test_version_option(runner, version_switch):
    """Test that --version option displays the version"""
    result = runner.invoke(cli.run, [version_switch])
    out = result.output.strip()

    assert result.exit_code == 0
    assert len(out.split('\n')) == 1
    assert 'version' in out


def test_choose_pkg_cmd_does_not_set(recipe_arguments):
    option_dict = {
        'pkg_cmd': 'testpkg',
    }
    cli._choose_pkg_cmd(option_dict)
    assert option_dict['pkg_cmd'] == 'testpkg'


def test_choose_pkg_cmd_sets_default_value(recipe_arguments):
    option_dict = {
        'pkg_cmd': None,
    }
    cli._choose_pkg_cmd(option_dict)
    assert option_dict['pkg_cmd'] == 'fedpkg'


@pytest.mark.parametrize('download_type', ('fedpkg', 'rhpkg'))
def test_choose_pkg_cmd_sets_download_type(recipe_arguments,
                                           download_type):
    option_dict = {
        'download': download_type,
        'pkg_cmd': None,
    }
    cli._choose_pkg_cmd(option_dict)
    assert option_dict['pkg_cmd'] == download_type


def test_choose_pkg_cmd_returns_optoin_dict_is_none():
    cli._choose_pkg_cmd(None)
    assert True


@pytest.mark.parametrize('value,exception_type', [
    (1, None),
    (1000, None),
    (-1, click.BadParameter),
])
def test_retry_sets_num_of_retries(value, exception_type, recipe_arguments):
    """Ensure proper handling of retry parameter."""

    options = ['--retry', str(value)]

    with ExitStack() as contexts:
        if exception_type:
            contexts.enter_context(pytest.raises(exception_type))

        ctx = contexts.enter_context(cli.run.make_context(
            'test-retry[{}]'.format(value),
            options + recipe_arguments,
        ))

        assert ctx.params['retry'] == value



"""Builder interface for the Koji build service."""

import logging
import re
from pathlib import Path
from typing import Iterator, Optional, Sequence

from ..utils import run_cmd, run_cmd_with_capture
from ..work import Work
from .base import BaseBuilder

LOG = logging.getLogger(__name__)


class KojiBuilder(BaseBuilder):
    """Builder interface for the Koji build service."""

    DEFAULT_TARGET_FORMAT = 'sclo{epel}-{collection}-rh-el{epel}'

    def __init__(
        self,
        work: Work,
        *,
        koji_scratch: bool = False,
        koji_epel: int = None,
        koji_target_format: str = None,
        koji_owner: str = None,
        koji_profile: Optional[str] = None,
        **extra_options
    ):
        """Initialize the builder.

        Keyword arguments:
            work: Overview of the work to do.
            koji_scratch: Indicates if the --scratch build option
                should be used.
            koji_epel: Version number of the EPEL to build for.
            koji_target_format: Format string for the target to build into.
            koji_owner: The user name of the owner of packages
                newly added to a build tag.
            koji_profile: Name of the configuration profile to use.
        """

        # Validity checks
        if koji_epel is None:
            raise ValueError('koji_epel parameter is required.')
        if koji_owner is None:
            raise ValueError('koji_owner parameter is required.')

        #: The collection to build
        self.collection = work._recipe._collection_id

        #: EPEL version to build for
        self.epel = koji_epel

        #: Format string for the build target
        self.target_format = koji_target_format or self.DEFAULT_TARGET_FORMAT

        #: Owner of packages added by this builder into a tag
        self.owner = koji_owner

        #: The configuration profile to use
        self.profile = koji_profile

        #: Scratch build indicator
        self.scratch_build = koji_scratch

        # Private target destination cache
        self._destination_tag_cache = {}

    @property
    def base_command(self) -> Sequence[str]:
        """Basis of the command for build service interaction."""

        binary = 'koji'

        if self.profile is None:
            return [binary]
        else:
            return [binary, '--profile', self.profile]

    @property
    def target(self) -> str:
        """The koji target to build into."""

        return self.target_format.format(
            collection=self.collection,
            epel=self.epel,
        )

    @staticmethod
    def adjust_bootstrap_release(
        source: Iterator[str],
        position: Optional[int]
    ):
        """Adjust Release value for bootstrapping, if necessary.

        Keyword arguments:
            source: The source file iterator.
            position: The position in the bootstrap sequence.

        Yields:
            Modified lines.
        """

        # Pass source if not bootstrapping
        if position is None:
            return source

        dist_original = re.compile(r'%\{?\??dist\}?')
        dist_replaced = r'.bs{:02}\g<0>'.format(position)

        for line in source:
            if line.startswith('Release:'):
                line = dist_original.sub(dist_replaced, line)

            yield line

    def prepare_extra_steps(self, source, package_dict):
        """Koji-specific SPEC adjustments."""

        # Add .bsXX to release tag on bootstrap
        contents = self.adjust_bootstrap_release(
            source=source,
            position=package_dict['bootstrap_position'],
        )

        return contents

    def build(self, package_dict, **kwargs):
        """Build a package using Koji instance"""

        srpm_path = self._make_srpm(
            name=package_dict['name'],
            collection=self.collection,
            epel=self.epel,
        )

        self._add_package(name='{collection}-{name}'.format(
            collection=self.collection,
            name=package_dict['name'],
        ))
        self._submit_build(srpm_path)
        self._wait_for_repo()

    @staticmethod
    def _make_srpm(name: str, collection: str, epel: int) -> Path:
        """Create SRPM of the specified name in current directory.

        Keyword arguments:
            name: Name of the package to create.
            collection: Name/identification of the package's collection.
            epel: The EPEL version to build for.

        Returns:
            Path to the created SRPM.
        """

        spec_path = Path('.'.join((name, 'spec')))
        if not spec_path.exists():
            raise FileNotFoundError(spec_path)

        # Mimic rpkg behaviour
        directory_kinds = 'source', 'spec', 'build', 'srcrpm', 'rpm', 'top'
        cwd = Path.cwd()

        # directories
        define_list = [
            '_{kind}dir {cwd}'.format(kind=k, cwd=cwd) for k in directory_kinds
        ]
        # dist tag
        define_list.append('dist .el{epel}'.format(epel=epel))
        # collection name – needed to generate prefixed package
        define_list.append('scl {collection}'.format(collection=collection))

        # Assemble the command
        command = ['rpmbuild']
        command += ['--define="{d}"'.format(d=d) for d in define_list]
        command += ['-bs', str(spec_path)]

        run_cmd_with_capture(' '.join(command))

        if name == collection:  # metapackage build
            glob_format = '{collection}-*.src.rpm'
        else:
            glob_format = '{collection}-{name}-*.src.rpm'

        srpm_path, = cwd.glob(glob_format.format(
            collection=collection,
            name=name,
        ))
        return srpm_path

    @property
    def _destination_tag(self) -> str:
        """Queries the destination tag for a current target."""

        # Destination tag is dependent on target name and profile
        # Note: extract to generic utility decorator?
        cache_key = self.profile, self.target

        cached = self._destination_tag_cache.get(cache_key, None)
        if cached is not None:
            return cached

        query = ['list-targets', '--name={self.target}'.format(self=self)]

        # FIXME: ugly parsing of koji output
        raw_output = run_cmd_with_capture(' '.join(self.base_command + query))
        target_line = raw_output.stdout.splitlines()[2]
        destination_tag = target_line.split()[-1].decode('utf-8')

        LOG.debug('Destination tag: {}'.format(destination_tag))

        self._destination_tag_cache[cache_key] = destination_tag
        return destination_tag

    def _add_package(self, name: str) -> None:
        """Add package to the destination tag for the target.

        Keyword arguments:
            name: Name of the package to add.
        """

        # Command is safely ignored if the package is in the tag already
        command = [
            'add-pkg',
            '--owner={owner}'.format(owner=self.owner),
            self._destination_tag,
            name,
        ]
        run_cmd(' '.join(self.base_command + command))

    def _submit_build(self, srpm_path: Path) -> None:
        """Submit build to the build service.

        Keyword arguments:
            srpm_path: Path to the SRPM to build.
            scratch_build: If true, build with scratch option
        """

        command = self.base_command + ['build']

        if self.scratch_build:
            command.append('--scratch')

        arguments = [self.target, str(srpm_path)]

        run_cmd(' '.join(command + arguments))

    def _wait_for_repo(self) -> None:
        """Wait for the package to appear in build repository."""

        # Packages are not submitted to repository on scratch builds
        if self.scratch_build:
            return

        command = self.base_command + ['wait-repo', '--target', self.target]
        run_cmd(' '.join(command))



import logging

import pytest

import rpmlb.logging


@pytest.fixture
def root_logger():

    log = logging.getLogger()

    default_level = log.getEffectiveLevel()

    try:
        yield log
    finally:
        log.setLevel(default_level)


def test_configure_logging_logs_on_verbose_false(root_logger):
    rpmlb.logging.configure_logging(False)
    assert root_logger.getEffectiveLevel() == logging.INFO


def test_configure_logging_logs_on_verbose_true(root_logger):
    rpmlb.logging.configure_logging(True)
    assert root_logger.getEffectiveLevel() == logging.DEBUG



from unittest import mock

import pytest

from rpmlb.builder.copr import CoprBuilder
from rpmlb.recipe import Recipe
from rpmlb.work import Work


@pytest.fixture
def work(valid_recipe_path):
    """Provide Work instance."""

    valid_recipe = Recipe(str(valid_recipe_path), 'rh-ror50')
    return Work(valid_recipe)


@pytest.fixture
def builder(work):
    pkg_cmd = 'fedpkg'
    copr_repo = 'ror50-test'
    return CoprBuilder(work, pkg_cmd=pkg_cmd, copr_repo=copr_repo)


def test_init(builder):
    assert builder


def test_init_raises_error_without_copr_repo(builder):
    pkg_cmd = 'fedpkg'
    with pytest.raises(ValueError):
        CoprBuilder(work, pkg_cmd=pkg_cmd)


def test_build_passes(builder, work):
    with mock.patch('rpmlb.utils.run_cmd',
                    mock.Mock(return_value=True)) as mock_run_cmd:
        package_dict = {'name': 'a'}
        builder.build(package_dict)
        mock_run_cmd.assert_any_call('rm -v *.rpm', check=False)
        mock_run_cmd.assert_any_call('fedpkg srpm')
        mock_run_cmd.assert_any_call('copr-cli build ror50-test *.rpm')
    assert True



from pathlib import Path
from unittest import mock

import pytest

from rpmlb.downloader.base import BaseDownloader


@pytest.fixture
def downloader():
    downloader = BaseDownloader()
    return downloader


def test_init(downloader):
    assert downloader


def test_run_calls_before_build_after(downloader):
    downloader.before = mock.MagicMock()
    downloader.download = mock.MagicMock(return_value=True)
    downloader.after = mock.MagicMock()

    mock_work = get_mock_work()
    downloader.run(mock_work)
    assert downloader.before.called
    assert downloader.download.called
    assert downloader.after.called


def test_run_returns_true_on_success(downloader):
    downloader.download = lambda *args, **kwargs: None

    mock_work = get_mock_work()
    assert downloader.run(mock_work)


def test_run_does_not_call_before_build_after_with_resume(downloader):
    downloader.before = mock.MagicMock()
    downloader.download = mock.MagicMock(return_value=True)
    downloader.after = mock.MagicMock()

    mock_work = get_mock_work()
    downloader.run(mock_work, resume=2)
    assert not downloader.before.called
    assert not downloader.download.called
    assert not downloader.after.called


def test_run_skip_download(downloader):
    downloader.download = mock.MagicMock(return_value=True)
    downloader._is_download_skipped = lambda *args, **kwargs: True

    mock_work = get_mock_work()
    with pytest.helpers.generate_tmp_path() as tmp_path:
        tmp_path.mkdir()
        with pytest.helpers.pushd(tmp_path):
            downloader.run(mock_work)
            assert not downloader.download.called
            assert Path('a').is_dir()
            assert Path('b').is_dir()


def test_is_download_skipped_raises_error_when_package_dict_none(downloader):
    package_dict = None
    with pytest.raises(ValueError):
        downloader._is_download_skipped(package_dict)


def test_is_download_skipped_returns_false_on_matched_dist(downloader):
    package_dict = {
        'name': 'a',
        'dist': 'fc2[56]',
    }
    kwargs = {
        'dist': 'fc26',
    }
    assert not downloader._is_download_skipped(package_dict, **kwargs)


def test_is_download_skipped_returns_true_on_unmatched_dist(downloader):
    package_dict = {
        'name': 'a',
        'dist': 'fc2[56]',
    }
    kwargs = {
        'dist': 'fc2',
    }
    assert downloader._is_download_skipped(package_dict, **kwargs)


def get_mock_work():
    mock_work = mock.MagicMock()
    package_dicts = [
        {'name': 'a'},
        {'name': 'b'},
    ]
    num_names = [
        '1',
        '2',
    ]
    mock_work.each_num_dir.return_value = iter(
        zip(package_dicts, num_names))
    return mock_work



import setuptools

with open('README.rst') as f:
    readme = f.read().strip()


setuptools.setup(
    name='rpmlb',
    license='GPLv2+',
    use_scm_version=True,
    description='RPM List Builder',
    long_description=readme,
    author='Jun Aruga',
    author_email='jaruga@redhat.com',
    url='https://github.com/sclorg/rpm-list-builder',
    packages=[
        'rpmlb',
        'rpmlb.builder',
        'rpmlb.downloader',
    ],
    install_requires=[
        'PyYAML',
        'retry',
        'click',
        'typing;python_version<"3.5"',
    ],
    entry_points={
        'console_scripts': [
            'rpmlb=rpmlb.cli:run',
        ]
    },
    setup_requires=[
        'pytest-runner',
        'setuptools_scm',
    ],
    classifiers=[
        'License :: OSI Approved :: '
        'GNU General Public License v2 or later (GPLv2+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
)



import logging

from .. import utils
from ..work import Work
from .base import BaseBuilder

LOG = logging.getLogger(__name__)


class CoprBuilder(BaseBuilder):
    """A builder class for Copr."""

    def __init__(self, work: Work, copr_repo: str = None, **options):
        """Initialize the builder.

        Keyword arguments:
            work: The overview of the work to do.
            copr_repo: Name of the repository to build into.
        """

        super().__init__(work, copr_repo=copr_repo, **options)

        if copr_repo is None:
            raise ValueError('copr_repo is required.')

        #: Name of the repository to build into
        self.copr_repo = copr_repo

    def build(self, package_dict, **kwargs):
        utils.run_cmd('rm -v *.rpm', check=False)
        utils.run_cmd('{} srpm'.format(self._pkg_cmd))
        utils.run_cmd('copr-cli build {} *.rpm'.format(self.copr_repo))
